/*
OpenTHFD Version 3.1.0 (CPU + OpenMP)
*/

#define MAIN
#include "oth.h"
#undef MAIN

#include "oth_prototype.h"

static void args(int, char *[], int *, int *, int *, int *, char []);

int main(int argc, char *argv[])
{
	const char prog[] = "(CPU+OpenMP)";
	const char errfmt[] = "*** file %s open error.\n";
	char str[BUFSIZ];
	double cpu[4];
	int ierr = 0;
	FILE *fp_in = NULL, *fp_out = NULL, *fp_log = NULL;

	const int io = 1;

	// arguments
	runMode = 0;
	SCALING = 1;
	MATRIX = 1;
	Iseed = 0;
	int nthread = 1;
	int ev = 0;
	int geom = 0;
	int prompt = 0;
	char fn_in[BUFSIZ] = "";
	args(argc, argv, &nthread, &ev, &geom, &prompt, fn_in);
	//printf("%d %d %d %d %d %d %d %d %s\n", runMode, SCALING, MATRIX, Iseed, nthread, ev, geom, prompt, fn_in); fflush(stdout);

	// set number of threads
#ifdef _OPENMP
	omp_set_num_threads(nthread);
#endif

	// input
	if (io) {
		if ((fp_in = fopen(fn_in, "r")) == NULL) {
			printf(errfmt, fn_in);
			fflush(stdout);
			ierr = 1;
		}
		if (!ierr) {
			ierr = input(fp_in, (!geom && (runMode < 2)));
			fclose(fp_in);
		}
	}
	if (ierr) {
		if (prompt) getchar();
		exit(1);
	}

	// === solver ===

	if ((runMode == 0) || (runMode == 1)) {
		// cpu time
		cpu[0] = cputime();

		// logo
		if (io) {
			sprintf(str, "<<< %s %s Ver.%d.%d.%d >>>", PROGRAM, prog, VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD);
		}

		if (!geom) {
			// open log file
			if (io) {
				if ((fp_log = fopen(FN_log, "w")) == NULL) {
					printf(errfmt, FN_log);
					fflush(stdout);
					ierr = 1;
				}
			}
			if (ierr) {
				if (prompt) getchar();
				exit(1);
			}

			// monitor
			if (io) {
				monitor1(fp_log, str);
				sprintf(str, "Thread = %d", nthread);
				monitor1(fp_log, str);
			}
		}
		else {
			if (io) {
				printf("%s\n", str);
				fflush(stdout);
			}
		}

		// plot geometry 3d
		if (io) {
			plot3dGeom(ev);
		}

		// plot geometry only : exit
		if (geom) {
			exit(0);
		}

		// setup
		setupIndex();
		setupMaterial(0);
		setup();
		alloc_farfield();

		// monitor
		if (io) {
			monitor2(fp_log, 0, 0);
		}

		// solve
		cpu[1] = cputime();
		solve(fp_log, io);
		cpu[2] = cputime();

		// output

		if (io) {
			// input imepedanece
			if (NFeed) {
				zfeed();
				outputZfeed(fp_log);
			}

			// S-parameters
			if (NPoint) {
				spara();
				outputSpara(fp_log);
			}

			// cross section
			if (IPlanewave) {
				outputCross(fp_log);
			}

			// power loss (if exists)
			outputPloss(fp_log);

			// coupling
			if (NFeed && NPoint) {
				outputCoupling(fp_log);
			}

			// output files
			monitor3(fp_log, ev);

			// write oth.out
			if (runMode == 1) {
				if ((fp_out = fopen(FN_out, "wb")) == NULL) {
					printf(errfmt, FN_out);
					getchar();
					exit(1);
				}
				writeout(fp_out);
				fclose(fp_out);
			}
		}

		// cpu time
		cpu[3] = cputime();

		if (io) {
			// cpu time
			monitor4(fp_log, cpu);

			// close log file
			fclose(fp_log);
		}
	}

	// === post ===

	if (io && ((runMode == 0) || (runMode == 2))) {
		// read oth.out
		if (runMode == 2) {
			if ((fp_out = fopen(FN_out, "rb")) == NULL) {
				printf(errfmt, FN_out);
				getchar();
				exit(1);
			}
			readout(fp_out);
			fclose(fp_out);
			alloc_farfield();
		}

		// post process
		post(ev);
	}

	// prompt
	if (io && prompt) getchar();

	return 0;
}

static void args(int argc, char *argv[],
	int *nthread, int *ev, int *geom, int *prompt, char fn_in[])
{
	const char usage[] = "Usage : oth [-solver|-post] [-matrix|-nomatrix] [-n <thread>] [-ev] [-geom] [-prompt] <datafile>";

	if (argc < 2) {
		printf("%s\n", usage);
		exit(0);
	}

	while (--argc) {
		++argv;
		if      (!strcmp(*argv, "-solver")) {
			runMode = 1;
		}
		else if (!strcmp(*argv, "-post")) {
			runMode = 2;
		}
		else if (!strcmp(*argv, "-scaling")) {
			SCALING = 1;
		}
		else if (!strcmp(*argv, "-noscaling")) {
			SCALING = 0;
		}
		else if (!strcmp(*argv, "-matrix")) {
			MATRIX = 1;
		}
		else if (!strcmp(*argv, "-nomatrix")) {
			MATRIX = 0;
		}
		else if (!strcmp(*argv, "-n")) {
			if (--argc) {
				*nthread = atoi(*++argv);
				if (*nthread < 1) *nthread = 1;
			}
			else {
				break;
			}
		}
		else if (!strcmp(*argv, "-ev")) {
			*ev = 1;
		}
		else if (!strcmp(*argv, "-geom")) {
			*geom = 1;
		}
		else if (!strcmp(*argv, "-prompt")) {
			*prompt = 1;
		}
		else if (!strcmp(*argv, "-random")) {
			if (--argc) {
				Iseed = atoi(*++argv);
			}
			else {
				break;
			}
		}
		else if (!strcmp(*argv, "--help")) {
			printf("%s\n", usage);
			exit(0);
		}
		else {
			strcpy(fn_in, *argv);
		}
	}
}
